"""Client with base functionality."""
from typing import Any, Optional

import httpx

from forager_forward.common.exceptions import ForagerAPIError


class BaseClient(object):
    """Base functionality for client."""

    def __init__(self, api_key: str) -> None:
        """Initialize client."""
        self.api_key: str = api_key
        self.endpoint: str = 'https://api.hunter.io/v2/'

    def _perform_request(
        self,
        operation: str,
        method: str = 'get',
        raw: bool = False,
        **kwargs: Any,
    ) -> dict | httpx.Response:
        """Perform http request."""
        param_dict: dict = kwargs.get('param_dict', {})
        param_dict['api_key'] = self.api_key
        request = httpx.Request(
            method,
            '{domain}{operation}'.format(domain=self.endpoint, operation=operation),
            params=param_dict,
            json=kwargs.get('payload'),
            headers=kwargs.get('headers'),
        )
        with httpx.Client() as client:
            response: httpx.Response = client.send(request)
        if raw:
            return response
        some_data: Optional[dict] = response.json().get('data')
        if some_data is not None:
            return some_data
        raise ForagerAPIError(response.json())

    async def _aperform_request(
        self,
        operation: str,
        method: str = 'get',
        raw: bool = False,
        **kwargs: Any,
    ) -> dict | httpx.Response:
        """Perform async http request."""
        param_dict: dict = kwargs.get('param_dict', {})
        param_dict['api_key'] = self.api_key
        request = httpx.Request(
            method,
            '{domain}{operation}'.format(domain=self.endpoint, operation=operation),
            params=param_dict,
            json=kwargs.get('payload'),
            headers=kwargs.get('headers'),
        )
        async with httpx.AsyncClient() as client:
            response = await client.send(request)
        if raw:
            return response
        some_data: Optional[dict] = response.json().get('data')
        if some_data is not None:
            return some_data
        raise ForagerAPIError(response.json())



"""Client for Forager project."""
from __future__ import annotations

from forager_forward.app_clients.base import BaseClient
from forager_forward.app_clients.email_client import AsyncEmailClient, EmailClient


class Client(BaseClient, EmailClient, AsyncEmailClient):
    """Base functionality for client."""



"""Storage for Forager project."""
from __future__ import annotations

from typing import Any

from forager_forward.common.exceptions import ForagerKeyError
from forager_forward.common.validators import common_validators


class Storage(object):
    """Storage with methods to perform CRUD operations, Singlton."""

    _storage: dict = {}

    def __new__(cls, *args: Any, **kwargs: Any) -> Storage:
        """Create new instance, if it's None, otherwise use earlier created one."""
        if getattr(cls, 'instance', None) is None:
            cls.instance = super().__new__(cls, *args, **kwargs)
        return cls.instance

    @property
    def storage(self) -> dict:
        """Get local storage."""
        return self._storage

    def create(self, key: str, some_data: Any) -> None:
        """Save arbitrary some_data to storage."""
        common_validators.validate_str('storage_key', key)
        if key in self._storage:
            raise ForagerKeyError(
                'Key {key} already presents in storage. Use "update" to modify some_data.'.format(key=key),
            )
        self._storage[key] = some_data

    def read(self, key: str) -> Any:
        """Read value from storage by key."""
        common_validators.validate_str('storage_key', key)
        return self._storage.get(key)

    def update(self, key: str, some_data: Any) -> None:
        """Update key some_data."""
        common_validators.validate_str('storage_key', key)
        if key not in self._storage:
            raise ForagerKeyError(
                'key {key} is not in storage. Use "create" operation.'.format(key=key),
            )
        self._storage[key] = some_data

    def delete(self, key: str) -> Any:
        """Delete key some_data pair and return some_data."""
        common_validators.validate_str('storage_key', key)
        return self._storage.pop(key, None)



"""Exceptions for Forager project."""


class ForagerError(Exception):
    """General exception for project."""


class ArgumentError(ForagerError):
    """Error with not correct arguments."""


class ArgumentValidationError(ForagerError):
    """Validation error, wrong type etc."""


class ForagerAPIError(ForagerError):
    """API error."""


class ForagerKeyError(ForagerError):
    """Error, if key presents in storage."""



"""Service for email validation with result saving to storage."""
from typing import Optional

from forager_forward.common.exceptions import ArgumentValidationError
from forager_forward.common.storage import Storage
from forager_forward.common.validators import validators


class EmailValidationService(object):
    """Class email validation and crud result."""

    _storage: Storage = Storage()

    def create_email_record(self, email: str) -> Optional[bool]:
        """
        Create email record in storage.

        :param email: str Email to create record.
        :return: bool True, if operation was successfull, otherwise False..
        """
        try:
            for validator in validators['email']:
                validator('email', email)
        except ArgumentValidationError:
            self._storage.create(email, some_data=False)
            return False
        self._storage.create(email, some_data=True)
        return True

    def read_email_record(self, email: str) -> Optional[dict]:
        """
        Read email record from storage.

        :param email: str Email to retrieve info.
        :return: dict Email validation info or None.
        """
        return self._storage.read(email)

    def delete_email_record(self, email: str) -> None:
        """
        Delete email record from storage.

        :param email: str Email to delete.
        :return: None
        """
        self._storage.delete(email)



"""Utilities for Forager project."""
from __future__ import annotations

from typing import Any

from forager_forward.common.validators import special_validators, validators


def create_and_validate_params(operation_type: str, **kwargs: Any) -> dict:
    """
    Add params from keyword arguments.

    :param operation_type: str Name of request operation.
    :param kwargs: dict Key word arguments for particular operation.
    :return: dict Params for request.
    """
    special_validators.validate_arguments(operation_type, kwargs)
    param_dict: dict = {}
    for key, element in kwargs.items():
        if element is not None:
            for validator in validators[key]:   # type: ignore
                validator(key, element)
            param_dict[key] = element
    for validation_handler in validators['required_arguments']:
        validation_handler(operation_type, param_dict)
    return param_dict



"""Type for Forager project."""
from typing import Callable, TypeAlias, TypedDict

Dict_A: TypeAlias = dict[str, str | int]
Callable_A: TypeAlias = Callable[[str, Dict_A], None]
Callable_B: TypeAlias = Callable[[str, str], None]
Callable_C: TypeAlias = Callable[[str, int], None]
Tuple_A: TypeAlias = tuple[Callable_A, ...]
Tuple_B: TypeAlias = tuple[Callable_B, ...]
Tuple_C: TypeAlias = tuple[Callable_C, ...]


class ValidatorTypeDict(TypedDict, total=False):
    """Type for validators dict."""

    domain: Tuple_B
    company: Tuple_B
    limit: Tuple_C
    offset: Tuple_C
    type: Tuple_B
    seniority: Tuple_B
    department: Tuple_B
    required_field: Tuple_B
    first_name: Tuple_B
    last_name: Tuple_B
    full_name: Tuple_B
    max_duration: Tuple_C
    email: Tuple_B
    required_arguments: Tuple_A



"""Forager client initializer."""
from __future__ import annotations

from typing import Any, Optional

from forager_forward.app_clients.client import Client


class ClientInitializer(object):
    """Singleton class for client initialization and retrieving."""

    _client: Optional[Client] = None

    def __new__(cls, *args: Any, **kwargs: Any) -> ClientInitializer:
        """Create new instance, if it's None, otherwise use earlier created one."""
        if getattr(cls, 'instance', None) is None:
            cls.instance = super().__new__(cls, *args, **kwargs)
        return cls.instance

    def initialize_client(self, api_key: str) -> None:
        """Initialize client instance."""
        self._client = Client(api_key)

    @property
    def client(self) -> Optional[Client]:
        """Get client instance."""
        return self._client



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['forager_forward',
 'forager_forward.app_clients',
 'forager_forward.app_services',
 'forager_forward.common']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.26.0,<0.27.0']

setup_kwargs = {
    'name': 'forager-forward',
    'version': '0.1.1',
    'description': 'Hunter.io v2 api implementation',
    'long_description': '# Forager clients and services\n\nA python wrapper for the Hunter.io v2 api as a client.\nEmail validation service.\n\n## Installation\n\n### Requirements\n\n    - Python 3.10\n    - httpx\n\n### To install\n\n   pip install forager_forward==0.1.1\n\n## Usage\n\nService supports next method from Hunter.io v2 api\n\n    - domain_search (with async adomain_search)\n    - email_finder (with async aemail_finder)\n    - verify_email (with async averify_email)\n    - email_count (with async aemail_count)\n    \nAdditionally, service supports crud methods for locally storing data\n\n## How to use service\n\n### Import service and instantiate it once\n\n    from forager_forward.client_initializer import ClientInitializer\n\n    initializer = ClientInitializer()\n\n    initializer.initialize_client("api_key_got_from_hunter")\n\n\n    client = initializer.client\n\n\n### Once initialized somewhere in the code you can get instances in different places without additional initialization\n\n    client = ClientInitializer().client\n\n\n### Search addresses for a given domain\n\n    client.domain_search("www.brillion.com.ua")\n\n### Or pass company name\n\n    client.domain_search(company="Brillion", limit=20, seniority="junior")\n\n### Find email address\n\n    client.email_finder(compayny="pmr", full_name="Sergiy Petrov", raw=True)\n\n### Check email deliverabelity\n\n    client.email_verifier("a@a.com")\n\n### All data can be stored in Storage class instance. It has its own crud methods, and it is Singleton.\n\n    from forager_forward.common.storage import Storage\n\n    storage = Storage()\n\n    storage.create(some_key, some_value)\n\n    storage.update(some_key, some_value)\n\n    some_variable = storage.read(some_key)\n\n    storage.delete(some_key)\n\n### To validate emails and store validation result use email_validation_service.\n\n    from forager_forward.app_services.email_validation_service import EmailValidationService\n\n    email_validator = EmailValidationService()\n\n    email_validator.create_email_record("some_email@company.com")\n\n    email_validator.read("another@company.com")\n\n## Tests\n\n    To run test firstly you need to install test dependency, then run\n\n        pytest -cov\n',
    'author': 'victro-nuzhniy',
    'author_email': 'nuzhniyva@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/victor-nuzhniy/forager.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""Forager project validators."""
import re

from forager_forward.common.exceptions import ArgumentValidationError
from forager_forward.common.project_types import ValidatorTypeDict

operation_arguments = {
    'domain-search': {'domain', 'company', 'limit', 'offset', 'type', 'seniority', 'department', 'required_field'},
    'email-finder': {'domain', 'company', 'first_name', 'last_name', 'full_name', 'max_duration'},
    'email-verifier': {'email'},
    'email-count': {'domain', 'company', 'type'},
}


class CommonValidators(object):
    """General type validators."""

    def validate_str(self, key: str, param_value: str) -> None:
        """Validate param_value is str type."""
        if not isinstance(param_value, str):
            raise ArgumentValidationError('{key} has wrong type.'.format(key=key))

    def validate_int(self, key: str, param_value: int) -> None:
        """Validate param_value is int type."""
        if not isinstance(param_value, int):
            raise ArgumentValidationError('{key} has wrong type.'.format(key=key))

    def validate_email(self, key: str, param_value: str) -> None:
        """Validate email."""
        specials = "!#$%&'*+-/=?^_`{|?."
        specials = re.escape(specials)
        regex = re.compile(
            '^(?!['
            + specials
            + '])(?!.*['
            + specials
            + ']{2})(?!.*['
            + specials
            + ']$)[A-Za-z0-9'
            + specials
            + ']+(?<!['
            + specials
            + '])@[A-Za-z0-9.-]+[.][A-Za-z]{2,4}$',
        )
        if not re.fullmatch(regex, param_value):
            raise ArgumentValidationError('{key} has invalid value.'.format(key=key))


common_validators = CommonValidators()


class KwargsValidators(object):
    """Kwargs validators."""

    def validate_email_type(self, key: str, param_value: str) -> None:
        """Validate email type."""
        if param_value not in {'personal', 'generic'}:
            raise ArgumentValidationError('{key} has wrong value.'.format(key=key))

    def validate_seniority(self, key: str, param_value: str) -> None:
        """Validate seniority kwarg."""
        value_list: list = param_value.split(', ')
        for elem in value_list:
            if elem not in {'junior', 'senior', 'executive'}:
                raise ArgumentValidationError('{key} has wrong value.'.format(key=key))

    def validate_department(self, key: str, param_value: str) -> None:
        """Validate department param_value."""
        value_list: list = param_value.split(', ')
        sample_set: set = {
            'executive',
            'it',
            'finance',
            'management',
            'sales',
            'legal',
            'support',
            'hr',
            'marketing',
            'communication',
        }
        for elem in value_list:
            if elem not in sample_set:
                raise ArgumentValidationError('{key} has wrong value.'.format(key=key))

    def validate_required_field(self, key: str, param_value: str) -> None:
        """Validate required field argument."""
        value_list: list = param_value.split(', ')
        for elem in value_list:
            if elem not in {'full_name', 'position', 'phone_number'}:
                raise ArgumentValidationError('{key} has wrong value.'.format(key=key))

    def validate_max_duration(self, key: str, param_value: int) -> None:
        """Validate max duration param_value."""
        high_limit: int = 20
        low_limit: int = 3
        if high_limit < param_value or param_value < low_limit:
            raise ArgumentValidationError('{key} should be in range from 3 to 20.'.format(key=key))


kwargs_validators = KwargsValidators()


class SpecialValidators(object):
    """Special case validators."""

    def validate_arguments(self, operation: str, arguments_dict: dict) -> None:
        """Validate params in allowed list for the operation."""
        arguments_set: set[str] | None = operation_arguments.get(operation)
        if arguments_set is None:
            raise ArgumentValidationError('{op} is not allowed operation'.format(op=operation))
        for key in arguments_dict:
            if key not in arguments_set:
                raise ArgumentValidationError(
                    'Argument {arg} is not from arguments list for {op} operation'.format(arg=key, op=operation),
                )

    def validate_required_arguments(self, key: str, param_dict: dict) -> None:
        """Validate presence required argument in the list."""
        if key in {'domain-search', 'email-finder', 'email-count'}:
            if 'domain' not in param_dict and 'company' not in param_dict:
                raise ArgumentValidationError(
                    'For {key} operation should be defined domain or company'.format(key=key),
                )

    def validate_email_finder_required_argument(self, key: str, param_dict: dict) -> None:
        """Validate precense at least 'full_name' or 'first_name' and 'last_name'."""
        if key == 'email-finder':
            if 'full_name' not in param_dict and ('first_name' not in param_dict or 'last_name' not in param_dict):
                raise ArgumentValidationError(
                    'At least first_name and last_name or full_name should be in param_dict.',
                )


special_validators = SpecialValidators()


validators: ValidatorTypeDict = {
    'domain': (common_validators.validate_str,),
    'company': (common_validators.validate_str,),
    'limit': (common_validators.validate_int,),
    'offset': (common_validators.validate_int,),
    'type': (common_validators.validate_str, kwargs_validators.validate_email_type),
    'seniority': (common_validators.validate_str, kwargs_validators.validate_seniority),
    'department': (common_validators.validate_str, kwargs_validators.validate_department),
    'required_field': (common_validators.validate_str, kwargs_validators.validate_required_field),
    'first_name': (common_validators.validate_str,),
    'last_name': (common_validators.validate_str,),
    'full_name': (common_validators.validate_str,),
    'max_duration': (common_validators.validate_int, kwargs_validators.validate_max_duration),
    'email': (common_validators.validate_str, common_validators.validate_email),
    'required_arguments': (
        special_validators.validate_required_arguments,
        special_validators.validate_email_finder_required_argument,
    ),
}



"""Email client for wrapping Hunter.io API."""
from abc import abstractmethod
from typing import Any, Optional

import httpx

from forager_forward.common.common_utilities import create_and_validate_params


class EmailClient(object):
    """Client for performing api calls."""

    def domain_search(
        self,
        domain: Optional[str] = None,
        company: Optional[str] = None,
        raw: bool = False,
        **kwargs: Any,
    ) -> dict | httpx.Response:
        """
        Perform domain_research request. Return all found email addresses.

        :param domain: str The domain on which to search for emails. Must be defined if company is not.
        :param company: str The name of the company on which to search for emails. Must be defined if domain is not.
        :param raw: bool Gives back the entire response instead of just the 'data'.
        :param kwargs: Any Can be from the list below:
            - limit: int The maximum number of emails to give back. Default is 10.
            - offset: int The number of emails to skip. Default is 0.
            - email_type: str The type of emails to give back. Can be one of 'personal' or 'generic'.
            - seniority: str The seniority level of the owners of emails to give back. Can be 'junior', 'senior',
            'executive' or a combination of them delimited by a comma.
            - department: str The department where the owners of the emails to give back work. Can be 'executive',
            'it', 'finance', 'management', 'sales', 'legal', 'support', 'hr', 'marketing', 'communication' or a
            combination of them delimited by a comma.
            - required_field: str Get only email addresses that have the selected field(s). The possible values
            are 'full_name', 'position', 'phone_number'. Several fields can be selected (comma-delimited).

        :return: Full payload of the query as a dict, with email addresses found.
        """
        operation: str = 'domain-search'
        param_dict: dict = create_and_validate_params(operation, domain=domain, company=company, **kwargs)
        return self._perform_request(operation, param_dict=param_dict, raw=raw)

    def email_finder(
        self,
        domain: Optional[str] = None,
        company: Optional[str] = None,
        raw: bool = False,
        **kwargs: Any,
    ) -> dict | httpx.Response:
        """
        Find the most likely email address from a domain name, first and a last name.

        :param domain: str The domain on which to search for emails. Must be defined if company is not.
        :param company: str The name of the company on which to search for emails. Must be defined if domain is not.
        :param raw: bool Gives back the entire response instead of just the 'data'.
        :param kwargs: Any Can be from the list below:
            - first_name: str The person's first name. It doesn't need to be in lowercase.
            - last_name: str The person's last name. It doesn't need to be in lowercase.
            - full_name: str The person's full name. Note that you'll get better results by supplying the person's
            first and last name if you can. It doesn't need to be in lowercase.
            - max_duration: int The maximum duration of the request in seconds.
            Setting a longer duration allows us to refine the results and provide more
            accurate data. It must range between 3 and 20. The default is 10.
        :return: Full payload of the query as a dict, with email addresses found.
        """
        operation: str = 'email-finder'
        param_dict: dict = create_and_validate_params(
            operation,
            domain=domain,
            company=company,
            **kwargs,
        )
        return self._perform_request(operation, param_dict=param_dict, raw=raw)

    def verify_email(
        self,
        email: str,
        raw: bool = False,
    ) -> dict | httpx.Response:
        """
        Verify the deliverability of an email address.

        :param email: str Email to verify.
        :param raw: bool Gives back the entire response instead of just the 'data'.
        :return: Full payload of the query as a dict.
        """
        operation: str = 'email-verifier'
        param_dict: dict = create_and_validate_params(
            operation,
            email=email,
        )
        return self._perform_request(operation, param_dict=param_dict, raw=raw)

    def email_count(
        self,
        domain: Optional[str] = None,
        company: Optional[str] = None,
        email_type: Optional[str] = None,
        raw: bool = False,
    ) -> dict | httpx.Response:
        """
        Count emails for domain or company.

        :param domain: str The domain on which to search for emails. Must be defined if company is not.
        :param company: str The name of the company on which to search for emails. Must be defined if domain is not.
        :param email_type: str The type of emails to give back. Can be one of 'personal' or 'generic'.
        :param raw: bool Gives back the entire response instead of just the 'data'.
        :return: Full payload of the query as a dict.
        """
        operation: str = 'email-count'
        param_dict: dict = create_and_validate_params(
            operation,
            domain=domain,
            company=company,
            type=email_type,
        )
        return self._perform_request(operation, param_dict=param_dict, raw=raw)

    @abstractmethod
    def _perform_request(
        self,
        operation: str,
        method: str = 'get',
        raw: bool = False,
        **kwargs: Any,
    ) -> dict | httpx.Response:
        """Perform http request."""


class AsyncEmailClient(object):
    """Client for performing async api calls."""

    async def adomain_search(
        self,
        domain: Optional[str] = None,
        company: Optional[str] = None,
        raw: bool = False,
        **kwargs: Any,
    ) -> dict | httpx.Response:
        """
        Perform domain_research request. Return all found email addresses.

        :param domain: str The domain on which to search for emails. Must be defined if company is not.
        :param company: str The name of the company on which to search for emails. Must be defined if domain is not.
        :param raw: bool Gives back the entire response instead of just the 'data'.
        :param kwargs: Any Can be from the list below:
            - limit: int The maximum number of emails to give back. Default is 10.
            - offset: int The number of emails to skip. Default is 0.
            - email_type: str The type of emails to give back. Can be one of 'personal' or 'generic'.
            - seniority: str The seniority level of the owners of emails to give back. Can be 'junior', 'senior',
            'executive' or a combination of them delimited by a comma.
            - department: str The department where the owners of the emails to give back work. Can be 'executive',
            'it', 'finance', 'management', 'sales', 'legal', 'support', 'hr', 'marketing', 'communication' or a
            combination of them delimited by a comma.
            - required_field: str Get only email addresses that have the selected field(s). The possible values
            are 'full_name', 'position', 'phone_number'. Several fields can be selected (comma-delimited).

        :return: Full payload of the query as a dict, with email addresses found.
        """
        operation: str = 'domain-search'
        param_dict: dict = create_and_validate_params(
            operation,
            domain=domain,
            company=company,
            **kwargs,
        )
        return await self._aperform_request(operation, param_dict=param_dict, raw=raw)

    async def aemail_finder(
        self,
        domain: Optional[str] = None,
        company: Optional[str] = None,
        raw: bool = False,
        **kwargs: Any,
    ) -> dict | httpx.Response:
        """
        Find the most likely email address from a domain name, first and a last name.

        :param domain: str The domain on which to search for emails. Must be defined if company is not.
        :param company: str The name of the company on which to search for emails. Must be defined if domain is not.
        :param raw: bool Gives back the entire response instead of just the 'data'.
        :param kwargs: Any Can be from the list below:
            - first_name: str The person's first name. It doesn't need to be in lowercase.
            - last_name: str The person's last name. It doesn't need to be in lowercase.
            - full_name: str The person's full name. Note that you'll get better results by supplying the person's
            first and last name if you can. It doesn't need to be in lowercase.
            - max_duration: int The maximum duration of the request in seconds.
            Setting a longer duration allows us to refine the results and provide more
            accurate data. It must range between 3 and 20. The default is 10.
        :return: Full payload of the query as a dict, with email addresses found.
        """
        operation: str = 'email-finder'
        param_dict: dict = create_and_validate_params(
            operation,
            domain=domain,
            company=company,
            **kwargs,
        )
        return await self._aperform_request(operation, param_dict=param_dict, raw=raw)

    async def averify_email(
        self,
        email: str,
        raw: bool = False,
    ) -> dict | httpx.Response:
        """
        Verify the deliverability of an email address.

        :param email: str Email to verify.
        :param raw: bool Gives back the entire response instead of just the 'data'.
        :return: Full payload of the query as a dict.
        """
        operation: str = 'email-verifier'
        param_dict: dict = create_and_validate_params(
            operation,
            email=email,
        )
        return await self._aperform_request(operation, param_dict=param_dict, raw=raw)

    async def aemail_count(
        self,
        domain: Optional[str] = None,
        company: Optional[str] = None,
        email_type: Optional[str] = None,
        raw: bool = False,
    ) -> dict | httpx.Response:
        """
        Count emails for domain or company.

        :param domain: str The domain on which to search for emails. Must be defined if company is not.
        :param company: str The name of the company on which to search for emails. Must be defined if domain is not.
        :param email_type: str The type of emails to give back. Can be one of 'personal' or 'generic'.
        :param raw: bool Gives back the entire response instead of just the 'data'.
        :return: Full payload of the query as a dict.
        """
        operation: str = 'email-count'
        param_dict: dict = create_and_validate_params(
            operation,
            domain=domain,
            company=company,
            type=email_type,
        )
        return await self._aperform_request(operation, param_dict=param_dict, raw=raw)

    @abstractmethod
    async def _aperform_request(
        self,
        operation: str,
        method: str = 'get',
        raw: bool = False,
        **kwargs: Any,
    ) -> dict | httpx.Response:
        """Perform async http request."""
